unit ifdib;

{ Support pixel formats:
    RGB 1,2,4,8,16(RGB555),24,32bit
    BITFIELDS 16,32bit
    RLE4, RLE8
}

interface

uses Windows, ImgFilter;

type
  PMaxBitmapInfo = ^TMaxBitmapInfo;
  TMaxBitmapInfo = packed record
    bmiHeader: TBitmapInfoHeader;
    case Boolean of
      True: (MaskR, MaskG, MaskB: DWORD);
      False: (bmiColors: array[0..255+3] of TRGBQuad);
  end;

  TDibFilter = class(TImgFilter)
    constructor Create(Alloc: TImgFilterMemAlloc); override;
    function Test(Buf: PChar; BufSize: Cardinal): Boolean; override;
    procedure Decode(InBuf: PChar; InBufSize: Cardinal;
        var OutBuf: PChar; var OutBufSize: Cardinal; var ImgInfo: TImgInfo);
        override;
  end;

procedure DecodeDIB(InBits: PChar; InBitsSize: Cardinal; Info: TMaxBitmapInfo;
    OutBuf: Pchar; OutBufSize: Cardinal);

implementation

uses NLib;

type
  PWord = ^Word;
  PTriByte = ^TTriByte;
  TTriByte = array[0..2] of Byte;

const
  MinHeaderSize = SizeOf(TBitmapFileHeader)+SizeOf(TBitmapCoreHeader);

procedure CoreInfoToBitmapInfo(BmInfo: PBitmapInfo; BmCore: PBitmapCoreInfo);
var
  i: Integer;
begin
  with BmInfo^ do
  begin
    FillChar(bmiHeader, SizeOf(bmiHeader), 0);
    with bmiHeader do
    begin
      biSize := SizeOf(bmiHeader);
      biWidth := BmCore.bmciHeader.bcWidth;
      biHeight := BmCore.bmciHeader.bcHeight;
      biPlanes := BmCore.bmciHeader.bcPlanes;
      biBitCount := BmCore.bmciHeader.bcBitCount;
      biCompression := BI_RGB;
    end;
    if (bmiHeader.biBitCount = 1) or (bmiHeader.biBitCount = 4)
       or (bmiHeader.biBitCount = 8) then
    begin
      for i := 0 to 1 shl bmiHeader.biBitCount - 1 do
      begin
        bmiColors[i].rgbBlue := BmCore.bmciColors[i].rgbtBlue;
        bmiColors[i].rgbGreen := BmCore.bmciColors[i].rgbtGreen;
        bmiColors[i].rgbRed := BmCore.bmciColors[i].rgbtRed;
      end;
    end;
  end;
end;

procedure GetBitmapInfo(Buf: PChar; BufSize: Cardinal;
    var BitmapInfo: TMaxBitmapInfo);
var
  tmpbuf: array[0..4096-1] of Byte;
  size: Cardinal;
  info: PChar;
begin
  size := SizeOf(tmpbuf);
  FillChar(tmpbuf, size, 0);
  if BufSize < size then
    size := BufSize;
  Move(Buf^, tmpbuf, size);
  info := @tmpbuf[SizeOf(TBitmapFileHeader)];

  if PBitmapCoreInfo(info).bmciHeader.bcSize = SizeOf(TBitmapCoreHeader) then
    CoreInfoToBitmapInfo(@BitmapInfo, PBitmapCoreInfo(info))
  else begin
    Move(info^, BitmapInfo, SizeOf(BitmapInfo));
    if BitmapInfo.bmiHeader.biCompression = BI_BITFIELDS then
      Exit;
    if (SizeOf(TBitmapFileHeader)+BitmapInfo.bmiHeader.biSize+
        SizeOf(BitmapInfo.bmiColors)) < SizeOf(tmpbuf) then
      Move(info[BitmapInfo.bmiHeader.biSize], BitmapInfo.bmiColors[0],
          SizeOf(BitmapInfo.bmiColors));
  end;
end;

function CalcStride(Width: Integer; Bpp: Integer): Integer;
begin
  Result := ((Width * Bpp + 7) shr 3 + 3) and not 3;
end;

procedure DecodeCLUT1(Buf: PChar; EndBuf: PChar; const Info: TMaxBitmapInfo;
  Line: PTriByte);
var
  x, m: Cardinal;
begin
  for x := 0 to Info.bmiHeader.biWidth-1 do
  begin
    if Buf >= EndBuf then
      Exit;
    m := 7 - x and 7; // place of bit
    Line^ := PTriByte(@Info.bmiColors[Byte(Buf^) shr m and 1])^;
    if m = 0 then
      Inc(Buf);
    Inc(Line);
  end;
end;

procedure DecodeCLUT4(Buf: PChar; EndBuf: PChar; const Info: TMaxBitmapInfo;
  Line: PTriByte);
var
  x: Cardinal;
begin
  for x := 0 to Info.bmiHeader.biWidth-1 do
  begin
    if Buf >= EndBuf then
      Exit;
    Line^ := PTriByte(
      @Info.bmiColors[Byte(Buf^) shr ((1-x and 1) * 4) and $F])^;
    Inc(Line);
    Inc(Buf, x and 1);
  end;
end;

procedure DecodeCLUT8(Buf: PChar; EndBuf: PChar; const Info: TMaxBitmapInfo;
  Line: PTriByte);
var
  x: Cardinal;
begin
  for x := 0 to Info.bmiHeader.biWidth-1 do
  begin
    if Buf >= EndBuf then
      Exit;
    Line^ := PTriByte(@Info.bmiColors[Byte(Buf^)])^;
    Inc(Line);
    Inc(Buf);
  end;
end;

procedure DecodeRGB888(Buf: PChar; EndBuf: PChar; const Info: TMaxBitmapInfo;
  Line: PTriByte);
var
  x: Cardinal;
begin
  for x := 0 to Info.bmiHeader.biWidth-1 do
  begin
    if Buf > EndBuf-3 then
      Exit;
    Line^ := PTriByte(Buf)^;
    Inc(Line);
    Inc(Buf, 3);
  end;
end;

procedure DecodeRGB8888(Buf: PChar; EndBuf: PChar; const Info: TMaxBitmapInfo;
  Line: PTriByte);
var
  x: Cardinal;
begin
  for x := 0 to Info.bmiHeader.biWidth-1 do
  begin
    if Buf > EndBuf-4 then
      Exit;
    Line^ := PTriByte(@Buf[0])^;
    Inc(Line);
    Inc(Buf, 4);
  end;
end;

type
  TDecodeLine = procedure (Buf: PChar; EndBuf: PChar;
      const Info: TMaxBitmapInfo; Line: PTriByte);

procedure DecodeRGB(InBuf: PChar; InBufSize: Integer;
    const Info: TMaxBitmapInfo; OutBuf: PChar; OutBufSize: Integer;
    DecodeLine: TDecodeLine);
var
  inend, pout: PChar;
  instride, outstride, y, h: Integer;
begin
  with Info do
  begin
    outstride := CalcStride(bmiHeader.biWidth, 24);
    instride := CalcStride(bmiHeader.biWidth, bmiHeader.biBitCount);
    inend := @InBuf[InBufSize];
    h := Abs(bmiHeader.biHeight);

    if bmiheader.biHeight < 0 then // topdown
    begin
      pout := PChar(@OutBuf[OutBufSize]) - outstride;
      outstride := -outstride;
    end else
      pout := OutBuf;
  end;

  for y := 0 to h-1 do
  begin
    DecodeLine(InBuf, inend, Info, PTriByte(pout));
    Inc(InBuf, instride);
    Inc(pout, outstride);
  end;
end;

function CountBits(Value: Cardinal): Integer;
begin
  Result := 0;
  repeat
    Inc(Result, Value and 1);
    Value := Value shr 1
  until Value = 0;
end;

function GetShift(Value: Cardinal): Integer;
begin
  Result := 0;
  Value := not Value;
  repeat
    Inc(Result, Value and 1);
    Value := Value shr 1
  until Value and 1 = 0;
end;

function ExpandBit(Value: Cardinal; NBits: Integer): Cardinal;
begin
  //  43210(5bit) -> 43210432(8bit)
  // 543210(6bit) -> 54321054(8bit)
  Result := (Value shl (8-NBits)) or (Value shr (NBits-(8-NBits)));
end;

function ExpandBit0123(Value: Cardinal; NBits: Integer): Cardinal;
const
  //    (0bit) -> 00000000(8bit)
  //   0(1bit) -> 00000000(8bit)
  //  10(2bit) -> 10101010(8bit)
  // 210(3bit) -> 21021021(8bit)
  BitExpandTab: array[0..3, 0..7] of Integer = (
    ($00, $00, $00, $00, $00, $00, $00, $00), // 0bit
    ($00, $ff, $00, $00, $00, $00, $00, $00), // 1bit
    ($00, $55, $aa, $ff, $00, $00, $00, $00), // 2bit
    ($00, $24, $49, $6d, $92, $b6, $db, $ff)  // 3bit
    );
begin
  Result := BitExpandTab[NBits, Value];
end;

function OmitBit(Value: Cardinal; NBits: Integer): Cardinal;
begin
  Result := Value shr (NBits-8);
end;

procedure DecodeBF(InBuf: PChar; InBufSize: Integer;
  const Info: TMaxBitmapInfo; OutBuf: PChar; OutBufSize: Integer);
var
  x, y, w, h, instride, outstride: Integer;
  inend: PChar;
  pin16: PWord;
  pin32: PCardinal;
  pout: PTriByte;
  rb, rs, gb, gs, bb, bs: Integer;
  rc, gc, bc: function (Value: Cardinal; NBits: Integer): Cardinal;
begin
  with Info do
  begin
    inend := @InBuf[InBufSize];
    instride := CalcStride(bmiHeader.biWidth, bmiHeader.biBitCount);
    outstride := CalcStride(bmiHeader.biWidth, 24);
    if Info.bmiHeader.biHeight < 0 then // topdown
    begin
      OutBuf := PChar(@OutBuf[OutBufSize]) - outstride;
      outstride := -outstride;
    end;
    w := bmiHeader.biWidth;
    h := Abs(bmiHeader.biHeight);

    rb := CountBits(MaskR);
    rs := GetShift(MaskR);
    if rb < 4 then
      rc := ExpandBit0123
    else if rb < 9 then
      rc := ExpandBit
    else
      rc := OmitBit;

    gb := CountBits(MaskG);
    gs := GetShift(MaskG);
    if gb < 4 then
      gc := ExpandBit0123
    else if gb < 9 then
      gc := ExpandBit
    else
      gc := OmitBit;

    bb := CountBits(MaskB);
    bs := GetShift(MaskB);
    if bb < 4 then
      bc := ExpandBit0123
    else if bb < 9 then
      bc := ExpandBit
    else
      bc := OmitBit;

    if bmiHeader.biBitCount = 16 then
      for y := 0 to h-1 do
      begin
        pin16 := @InBuf[instride*y];
        pout := @OutBuf[outstride*y];
        for x := 0 to w-1 do
        begin
          if pin16 > inend-SizeOf(pin16^) then
            Exit;
          pout[0] := bc(pin16^ and MaskB shr bs, bb);
          pout[1] := gc(pin16^ and MaskG shr gs, gb);
          pout[2] := rc(pin16^ and MaskR shr rs, rb);
          Inc(pin16);
          Inc(pout);
        end;
      end
    else // biBitCount = 32
      for y := 0 to h-1 do
      begin
        pin32 := @InBuf[instride*y];
        pout := @OutBuf[outstride*y];
        for x := 0 to w-1 do
        begin
          if pin32 > inend-SizeOf(pin32^) then
            Exit;
          pout[0] := bc(pin32^ and MaskB shr bs, bb);
          pout[1] := gc(pin32^ and MaskG shr gs, gb);
          pout[2] := rc(pin32^ and MaskR shr rs, rb);
          Inc(pin32);
          Inc(pout);
        end;
      end
  end;
end;

procedure DecodeRLE4(InBuf: PChar; InBufSize: Integer;
  const Info: TMaxBitmapInfo; OutBuf: PChar; OutBufSize: Integer);
var
  x, y, c0, c1, outstride, i: Integer;
  pin, pinend, pout, poutend: PChar;
begin
  FillChar(OutBuf^, OutBufSize, 0);
  pin := InBuf;
  pinend := @InBuf[InBufSize];
  outstride := CalcStride(Info.bmiHeader.biWidth, 24);
  if Info.bmiHeader.biHeight < 0 then // topdown
  begin
    OutBuf := PChar(@OutBuf[OutBufSize]) - outstride;
    outstride := -outstride;
  end;
  pout := OutBuf;
  poutend := @OutBuf[OutBufSize];
  y := 0;

  while (@pin[1] < pinend) and (@pout[2] < poutend) do
  begin
    c0 := Byte(pin[0]);
    c1 := Byte(pin[1]);
    Inc(pin, 2);

    if c0 = 0 then
    begin
      case c1 of
        0: begin // end of line
            Inc(y);
            pout := @OutBuf[outstride*y];
           end;
        1: Exit; // end of bitmap
        2: begin // delta
             if @pin[1] >= pinend then
               Exit;
             x := Byte(pin[0]);
             y := Byte(pin[1]);
             Inc(pin, 2);
             Inc(pout, x*3+outstride*y);
           end;
      else // absolute mode
        for i := 0 to c1-1 do
        begin
          PTriByte(pout)^ := PTriByte(
              @Info.bmiColors[Byte(pin^) shr ((1-i and 1) * 4) and $F])^;
          Inc(pin, i and 1);
          Inc(pout, 3);
          if @pout[2] >= poutend then
            Exit;
        end;
        Inc(pin, c1 and 1 + (c1+1) shr 1 and 1); // word boundary
      end;
    end
    else begin // encoded mode
      for i := 0 to c0-1 do
      begin
        PTriByte(pout)^ := PTriByte(
            @Info.bmiColors[c1 shr ((1-i and 1) * 4) and $F])^;
        Inc(pout, 3);
        if @pout[2] >= poutend then
          Exit;
      end;
    end;
  end;
end;

procedure DecodeRLE8(InBuf: PChar; InBufSize: Integer;
  const Info: TMaxBitmapInfo; OutBuf: PChar; OutBufSize: Integer);
var
  x, y, c0, c1, outstride, i: Integer;
  pin, pinend, pout, poutend: PChar;
begin
  FillChar(OutBuf^, OutBufSize, 0);
  pin := InBuf;
  pinend := @InBuf[InBufSize];
  outstride := CalcStride(Info.bmiHeader.biWidth, 24);
  if Info.bmiHeader.biHeight < 0 then // topdown
  begin
    OutBuf := PChar(@OutBuf[OutBufSize]) - outstride;
    outstride := -outstride;
  end;
  pout := OutBuf;
  poutend := @OutBuf[OutBufSize];
  y := 0;

  while (@pin[1] < pinend) and (@pout[2] < poutend) do
  begin
    c0 := Byte(pin[0]);
    c1 := Byte(pin[1]);
    Inc(pin, 2);

    if c0 = 0 then
    begin
      case c1 of
        0: begin // end of line
            Inc(y);
            pout := @OutBuf[outstride*y];
           end;
        1: Exit; // end of bitmap
        2: begin // delta
             if @pin[1] >= pinend then
               Exit;
             x := Byte(pin[0]);
             y := Byte(pin[1]);
             Inc(pin, 2);
             Inc(pout, x*3+outstride*y);
           end;
      else // absolute mode
        for i := 0 to c1-1 do
        begin
          PTriByte(pout)^ :=  PTriByte(@Info.bmiColors[Byte(pin^)])^;
          Inc(pin);
          Inc(pout, 3);
          if @pout[2] >= poutend then
            Exit;
        end;
        Inc(pin, c1 and 1); // word boundary
      end;
    end
    else begin // encoded mode
      for i := 0 to c0-1 do
      begin
        PTriByte(pout)^ :=  PTriByte(@Info.bmiColors[c1])^;
        Inc(pout, 3);
        if @pout[2] >= poutend then
          Exit;
      end;
    end;
  end;
end;

procedure DecodeDIB(InBits: PChar; InBitsSize: Cardinal;
    Info: TMaxBitmapInfo; OutBuf: Pchar; OutBufSize: Cardinal);
var
  decodeline: TDecodeLine;
begin
  with Info.bmiHeader do
  begin
    case biCompression of
      BI_RGB:
        begin
          decodeline := nil;

          case biBitCount of
            1: decodeline := DecodeCLUT1;
            4: decodeline := DecodeCLUT4;
            8: decodeline := DecodeCLUT8;
            16: begin
                  // RGB555
                  Info.MaskR := $7C00;
                  Info.MaskG := $03E0;
                  Info.MaskB := $001F;
                  DecodeBF(InBits, InBitsSize, Info, OutBuf, OutBufSize);
                end;
            24: decodeline := DecodeRGB888;
            32: decodeline := DecodeRGB8888;
          else
            raise EImgFilterError.Create('invalid bit count');
          end;

          if Assigned(decodeline) then
            DecodeRGB(InBits, InBitsSize, Info, OutBuf, OutBufSize,
                decodeline);
        end;
      BI_BITFIELDS:
        case biBitCount of
          16, 32: DecodeBF(InBits, InBitsSize, Info, OutBuf, OutBufSize);
        else
          raise EImgFilterError.Create('invalid bit count');
        end;
      BI_RLE4:
        DecodeRLE4(InBits, InBitsSize, Info, OutBuf, OutBufSize);
      BI_RLE8:
        DecodeRLE8(InBits, InBitsSize, Info, OutBuf, OutBufSize);
    else
      raise EImgFilterError.Create('unknown compression');
    end;
  end;
end;

function GetFormatName(Buf: PChar; BufSize: Cardinal): string;
var
  bmi: PMaxBitmapInfo;
  s, v: string;
  rb, gb, bb, ab: Integer;
begin
  bmi := @Buf[SizeOf(TBitmapFileHeader)];

  if bmi.bmiHeader.biSize = SizeOf(TBitmapCoreHeader) then
  begin
    Result := Format('OS/2 Bitmap, {0}bit',
        [PBitmapCoreInfo(bmi).bmciHeader.bcBitCount]);
    Exit;
  end;

  case bmi.bmiHeader.biCompression of
    BI_RGB:
      if bmi.bmiHeader.biBitCount = 16 then
        s := 'RGB555'
      else
        s := 'RGB888';
    BI_RLE8: s := 'RLE8';
    BI_RLE4: s := 'RLE4';
    BI_BITFIELDS:
      begin
        rb := CountBits(bmi.MaskR);
        gb := CountBits(bmi.MaskG);
        bb := CountBits(bmi.MaskB);
        if (bmi.bmiHeader.biSize = SizeOf(TBitmapV4Header)) or
           (bmi.bmiHeader.biSize = SizeOf(TBitmapV5Header)) then
        begin
          ab := CountBits(PBitmapV4Header(@bmi).bV4AlphaMask);
          s := Format('BITFIELDS (RGBA{0}{1}{2}{3})', [rb, gb, bb, ab]);
        end else
          s := Format('BITFIELDS (RGB{0}{1}{2})', [rb, gb, bb]);
      end;
  else
    s := 'unknown';
  end;

  if bmi.bmiHeader.biHeight < 0 then
    s := s + ', topdown';

  if bmi.bmiHeader.biSize = SizeOf(TBitmapV4Header) then
    v := ' V4'
  else if bmi.bmiHeader.biSize = SizeOf(TBitmapV5Header) then
    v := ' V5'
  else
    v := '';

  Result := 'Windows Bitmap'+v+', '+s;
end;


{ TDibFilter }

constructor TDibFilter.Create(Alloc: TImgFilterMemAlloc);
begin
  inherited;
  FName := 'Internal DIB decoder';
end;

function TDibFilter.Test(Buf: PChar; BufSize: Cardinal): Boolean;
begin
  if BufSize < MinHeaderSize then
  begin
    Result := False;
    Exit;
  end;

  if not ((Buf[0] = 'B') and (Buf[1] = 'M')) then
  begin
    Result := False;
    Exit;
  end;

  with PBitmapInfoHeader(@Buf[SizeOf(TBitmapFileHeader)])^ do
    Result :=
      (bisize = SizeOf(TBitmapCoreHeader)) or
      (bisize = SizeOf(TBitmapInfoHeader)) or
      // biSize of BITFIELDS Bitmap which Photoshop Elements 2 output is 56.
      (bisize = SizeOf(TBitmapInfoHeader)+16) or
      (bisize = SizeOf(TBitmapV4Header)) or
      (bisize = SizeOf(TBitmapV5Header));
end;

procedure TDibFilter.Decode(InBuf: PChar; InBufSize: Cardinal;
    var OutBuf: PChar; var OutBufSize: Cardinal; var ImgInfo: TImgInfo);
var
  bmi: TMaxBitmapInfo;
begin
  if InBufSize < MinHeaderSize then
    raise EImgFilterError.Create('invalid header');

  GetBitmapInfo(InBuf, InBufSize, bmi);

  ImgInfo.Width := bmi.bmiHeader.biWidth;
  if bmi.bmiHeader.biHeight < 0 then
    ImgInfo.Height := -bmi.bmiHeader.biHeight
  else
    ImgInfo.Height := bmi.bmiHeader.biHeight;
  ImgInfo.Components := 3;
  ImgInfo.FormatName := GetFormatName(InBuf, InBufSize);
  ImgInfo.Bpp := bmi.bmiHeader.biBitCount;
  ImgInfo.Comment := '';

  OutBufSize := ((ImgInfo.Width * 3 + 3) and not 3) * ImgInfo.Height;
  OutBuf := Alloc(OutBufSize);
  if OutBuf = nil then
    raise EImgFilterError.Create('output buffer allocation failed');

  with PBitmapFileHeader(InBuf)^ do
    DecodeDIB(@InBuf[bfOffBits], InBufSize-bfOffBits, bmi, OutBuf, OutBufSize);
end;

end.
